# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import logging
import json

import appdirs

from .core import GAMES

DEFAULT_CONFIG_DIR = appdirs.user_config_dir("PyParadox")

_logger = logging.getLogger(__name__)


def default_values(game):
    """Create and return a dictionary containing the default configuration
    values for each supported platform (Windows/OS X/Linux).

    :param str game: Shorthand version of the game (i.e., ck2, eu4, stel).
    :return: Default configuration values for ``mods``, ``excluded_dlcs``,
             ``binary_path`` and ``mod_path``.
    :rtype: dict
    """
    mods = []
    excluded_dlcs = []

    full_game = GAMES[game]

    # Windows
    if sys.platform.startswith("win"):
        binary_path = os.path.join(os.environ["ProgramFiles(x86)"],
                                   "Steam",
                                   "SteamApps",
                                   "common",
                                   full_game,
                                   "{}.exe".format(game))
        mod_path = os.path.join(os.path.expanduser("~"),
                                "Documents",
                                "Paradox Interactive",
                                full_game,
                                "mod")
    # OS X
    elif sys.platform == 'darwin':
        binary_path = os.path.join(appdirs.user_data_dir("Steam"),
                                   "steamapps",
                                   "common",
                                   full_game,
                                   "{}.app".format(game))
        mod_path = os.path.join(os.path.expanduser("~"),
                                "Documents",
                                "Paradox Interactive",
                                full_game,
                                "mod")
    # Linux
    else:
        binary_path = os.path.join(os.path.expanduser("~"),
                                   ".steam",
                                   "root",
                                   "steamapps",
                                   "common",
                                   full_game,
                                   game)
        mod_path = os.path.join(os.path.expanduser("~"),
                                ".paradoxinteractive",
                                full_game,
                                "mod")

    return {
        "game": game,
        "mods": mods,
        "excluded_dlcs": excluded_dlcs,
        "binary_path": binary_path,
        "mod_path": mod_path
    }


class JsonBackend:
    """Backend that implements an interface for reading and writing a
    :class:`Config` to a file, using the json format.
    """

    @staticmethod
    def read(stream):
        """Gather json data from the stream and return a dictionary of the
        gathered data.

        :param stream: Stream that contains json data.
        :type stream: file-like object
        :return: Data gathered from *stream*.
        :rtype: dict
        """
        return json.load(stream)

    @staticmethod
    def write(data, stream):
        """Write the data to a stream in a json format.

        :param dict data: Serialisable data.
        :param stream: Writeable stream.
        :type stream: file-like object
        """
        json.dump(data, stream, indent=4, sort_keys=True)


class Config(dict):
    """Simple dictionary with the added ability to read and write to file
    streams using (mostly) any format.
    """

    def __init__(self, *args, **kwargs):
        """Constructor for :class:`dict`.

        :keyword backend: Object with *read* and *write* properties.
        """

        self.backend = kwargs.pop("backend", JsonBackend)
        super().__init__(*args, **kwargs)

    @staticmethod
    def build_config(stream_or_file, default=None, backend=JsonBackend):
        """Factory method for :class:`Config`.

        Accepts a stream or file as  argument, from which the contents will be
        forwarded to the constructor of the *Config* object.

        If *default* is provided, the contents of the *Config* object will be
        set to the provided default values if the file or stream is invalid.

        :param stream_or_file: Holds the contents that the *Config* object will
                               be initialised with.
        :type stream_or_file: file-like object or str
        :keyword dict default: Default values if reading from *stream_or_file*
                               was unsuccessful.
        :keyword backend: Backend that will be used to read *stream_or_file*,
                          and the *Config* object will be initialised with.
        :return: *Config* object.
        :rtype: :class:`Config`
        """
        try:
            stream = open(stream_or_file)
        except TypeError:
            stream = stream_or_file
        except FileNotFoundError:
            _logger.warning("{} not found, using default".format(
                stream_or_file))
            stream = None

        if stream:
            data = backend.read(stream)
            stream.close()
        else:
            data = default if default is not None else dict()

        return Config(data, backend=backend)

    def load(self, stream, override=True, suppress_merge=False):
        """Load the contents of *stream* to self.

        *override* determines whether the contents of *stream* should be able
        to override self.  If *override* is :const:`False` and a key is being
        overridden, it will raise an :class:`AttributeError`.

        *suppress_merge* is only relevant if *override* is :const:`False`.  It
        explicitly silences errors that arise if a key already exists in the
        dictionary.

        :param stream: Stream that contains a valid format for the backend to
                       read.
        :type stream: file-like object
        :keyword bool override: Whether keys from *stream* should be able to
                                override self.
        :keyword bool suppress_merge: Whether errors arising from conflicting
                                      keys should be silenced.
        :raise AttributeError: If a key from *stream* conflicts with self.
        """
        data = self.backend.read(stream)
        if override:
            self.update(data)
        else:
            for key, value in data.items():
                if key in self:
                    if not suppress_merge:
                        raise AttributeError("key '{}' already exists".format(
                            key))
                else:
                    self[key] = value

    def save(self, stream):
        """Save self to *stream*, using the backend format.

        :param stream: Writeable stream.
        :type stream: file-like object
        """
        _logger.info("saving config to {}".format(stream.name))
        self.backend.write(self, stream)


class FileConfig(Config):
    """Sub-class of :class:`Config` that internally stores a file path that
    it uses every time a ``load()`` or ``save()`` method is called.
    """

    def __init__(self, path, *args, **kwargs):
        """Constructor for :class:`FileConfig`.

        :param str path: Path where the configuration file can be found.
        """
        super().__init__(*args, **kwargs)
        self.path = path

    @staticmethod
    def build_config(path, default=None, backend=JsonBackend):
        """Factory method for :class:`FileConfig`. Most functionality exists in
        :meth:`Config.build_config`.

        :param str path: Path where the configuration file can be found.
        :param dict default: Default values if reading from *stream_or_file*
                             was unsuccessful.
        :param backend: Backend that will be used to read *stream_or_file*,
                        and the *Config* object will be initialised with.
        """
        config = Config.build_config(path, default=default, backend=backend)

        return FileConfig(path, dict(config.items()), backend=backend)

    def load(self, override=True, suppress_merge=True):
        """Load the contents of *path* onto self. Most logic contained within
        :meth:`Config.load``.

        :keyword bool override: Whether keys from *path* should be able to
                                override self.
        :keyword bool suppress_merge: Whether errors arising from conflicting
                                      keys should be silenced.
        :raise AttributeError: If a key from *path* conflicts with self.
        """
        with open(self.path):
            super().load(self.path, override=override,
                         suppress_merge=suppress_merge)

    def save(self):
        """Save self to *path*, using the backend format."""
        _mkdir_p(os.path.dirname(self.path))

        with open(self.path, "w") as fp:
            super().save(fp)


def _mkdir_p(directory):
    try:
        os.makedirs(directory)
    except FileExistsError:
        pass



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
from itertools import chain
import pkg_resources

from PyQt5 import QtCore, QtGui, QtWidgets, uic

from .core import Plugin, find_plugins, make_command, execute_command

_logger = logging.getLogger(__name__)


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, config, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(pkg_resources.resource_filename(
            "pyparadox",
            "resources/ui/MainWindow.ui"),
            self)

        self._config = config

        modsModel = PluginListModel(self)
        self.modsView.setModel(modsModel)
        dlcsModel = PluginListModel(self)
        self.dlcsView.setModel(dlcsModel)
        expansionsModel = PluginListModel(self)
        self.expansionsView.setModel(expansionsModel)

        self._logicWrapper = LogicWrapper(
            self._config,
            modsModel,
            dlcsModel,
            expansionsModel)

        self._logicWrapper.resetModels()

        image = QtGui.QPixmap(pkg_resources.resource_filename(
            "pyparadox",
            "resources/{}.png".format(self._config["game"])))
        self.imageLbl.setPixmap(image)

        self.setWindowIcon(QtGui.QIcon(pkg_resources.resource_filename(
                                       "pyparadox", "resources/paradox.png")))

        self._logicWrapper.runFailed.connect(self._recoverRunFailed)
        self._logicWrapper.runSucceeded.connect(self.close)

    @QtCore.pyqtSlot(str, str)
    def _recoverRunFailed(self, command, error):
        _logger.error("failed to launch {}".format(command))
        msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning,
                                       self.tr("Error"),
                                       self.tr("Game could not launch."),
                                       QtWidgets.QMessageBox.Ok,
                                       self)
        message = "\n\n".join([command, error])
        msgBox.setDetailedText(message)
        msgBox.setDefaultButton(QtWidgets.QMessageBox.Ok)
        msgBox.exec_()

    @QtCore.pyqtSlot(bool)
    def on_configBtn_clicked(self, pressed):
        configDlg = ConfigDlg(self)
        configDlg.binaryPathEdit.setText(self._config["binary_path"])
        configDlg.modPathEdit.setText(self._config["mod_path"])

        if configDlg.exec_():
            self._config["binary_path"] = configDlg.binaryPathEdit.text()
            self._config["mod_path"] = configDlg.modPathEdit.text()
            self._logicWrapper.savePlugins()
            self._logicWrapper.resetModels()

    @QtCore.pyqtSlot(bool)
    def on_runBtn_clicked(self, pressed):
        self._logicWrapper.savePlugins()

        self._logicWrapper.run()


class ConfigDlg(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(pkg_resources.resource_filename(
            "pyparadox",
            "resources/ui/ConfigDlg.ui"),
            self)

    @QtCore.pyqtSlot(bool)
    def on_binaryPathBtn_clicked(self, pressed):
        old_path = self.binaryPathEdit.text()
        if os.path.isfile(old_path):
            old_dir = os.path.dirname(old_path)
        else:
            old_dir = os.path.expanduser("~")

        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            self.tr("Choose binary"),
            old_dir)

        if path:
            self.binaryPathEdit.setText(path)

    @QtCore.pyqtSlot(bool)
    def on_modPathBtn_clicked(self, pressed):
        old_path = self.modPathEdit.text()
        if not os.path.isdir(old_path):
            old_path = os.path.expanduser("~")

        path = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            self.tr("Choose mod folder"),
            old_path)

        if path:
            self.modPathEdit.setText(path)


class PluginListModel(QtCore.QAbstractListModel):

    NameRole = QtCore.Qt.UserRole + 1

    _roles = {NameRole: b"name", QtCore.Qt.CheckStateRole: b"checkState"}

    def __init__(self, parent=None):
        super(PluginListModel, self).__init__(parent)
        self._plugins = []

    def __iter__(self):
        return iter(self._plugins)

    def setPluginList(self, plugins):
        self.beginResetModel()
        self._plugins = sorted(plugins, key=lambda x: x.name)
        self.endResetModel()

    def rowCount(self, parent):
        return len(self._plugins)

    def data(self, index, role):
        item = self._plugins[index.row()]
        if role in (QtCore.Qt.DisplayRole, self.NameRole):
            return item.name
        elif role == QtCore.Qt.CheckStateRole:
            return QtCore.Qt.Checked if item.enabled else QtCore.Qt.Unchecked

    def flags(self, index):
        return (QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable |
                QtCore.Qt.ItemIsSelectable)

    def setData(self, index, value, role):
        item = self._plugins[index.row()]
        if role == QtCore.Qt.CheckStateRole:
            _logger.debug("{} {}".format(
                "enabled" if value else "disabled",
                item.name
            ))
            item.enabled = bool(value)
            self.dataChanged.emit(index, index, [role])
        else:
            return False
        return True

    def roleNames(self):
        roles = super().roleNames()
        roles.update(self._roles)
        return roles


class LogicWrapper(QtCore.QObject):

    runFailed = QtCore.pyqtSignal(str, str)
    runSucceeded = QtCore.pyqtSignal()

    configBinaryPathChanged = QtCore.pyqtSignal()
    configModPathChanged = QtCore.pyqtSignal()

    def __init__(self, config, modsModel, dlcsModel, expansionsModel,
                 parent=None):
        super().__init__(parent)
        self._config = config
        self._modsModel = modsModel
        self._dlcsModel = dlcsModel
        self._expansionsModel = expansionsModel

    @QtCore.pyqtSlot()
    def resetModels(self):
        mods = find_plugins(self._config["mod_path"], "mod",
                            config=self._config)
        dlc_path = os.path.join(os.path.dirname(self._config["binary_path"]),
                                "dlc")
        dlc_likes = find_plugins(dlc_path, "dlc", config=self._config)
        dlcs = [dlc for dlc in dlc_likes if dlc.plugin_type == Plugin.DLC]
        expansions = [exp for exp in dlc_likes if exp.plugin_type ==
                      Plugin.EXPANSION]

        self._modsModel.setPluginList(mods)
        self._dlcsModel.setPluginList(dlcs)
        self._expansionsModel.setPluginList(expansions)

    @QtCore.pyqtSlot()
    def savePlugins(self):
        mods = [mod.file_name for mod in self._modsModel if mod.enabled]
        excluded_dlcs = [dlc.file_name for dlc in chain(self._dlcsModel,
                         self._expansionsModel) if not dlc.enabled]

        self._config["mods"] = mods
        self._config["excluded_dlcs"] = excluded_dlcs
        self._config.save()

    @QtCore.pyqtSlot()
    def run(self):
        command = make_command(self._config["binary_path"],
                               self._config["mods"],
                               self._config["excluded_dlcs"])
        try:
            execute_command(command)
        except OSError as e:
            self.runFailed.emit(" ".join(command), str(e))
        else:
            self.runSucceeded.emit()

    @QtCore.pyqtProperty(str, notify=configBinaryPathChanged)
    def configBinaryPath(self):
        return self._config["binary_path"]

    @configBinaryPath.setter
    def configBinaryPath(self, value):
        self._config["binary_path"] = value
        self.configBinaryPathChanged.emit()

    @QtCore.pyqtProperty(str, notify=configModPathChanged)
    def configModPath(self):
        return self._config["mod_path"]

    @configModPath.setter
    def configModPath(self, value):
        self._config["mod_path"] = value
        self.configModPathChanged.emit()



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import io
import re
import unittest
from unittest import mock
from unittest.mock import patch, mock_open

from pyparadox.config import Config, FileConfig, JsonBackend, default_values

_CURRENT_DIRECTORY = os.path.dirname(__file__)
_MOD_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-mod")
_GAME_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-game")
_CONFIG_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-config")
_CONFIG_PATH = os.path.join(_CONFIG_DIRECTORY, "pyparadox.json")


class TestConfigModule(unittest.TestCase):

    @patch("pyparadox.config.sys", autospec=True)
    @patch("pyparadox.config.os.environ", autospec=True)
    @patch("pyparadox.config.os.path.expanduser", autospec=True)
    def test_default_values_windows(self, config_expanduser, config_environ,
                                    config_sys):
        config_sys.configure_mock(platform="win32")
        envs = {"ProgramFiles(x86)": "/"}
        config_environ.__getitem__.side_effect = lambda key: envs[key]
        config_expanduser.return_value = "/home"

        values = default_values("ck2")
        self.assertEqual(values["mods"], [])
        self.assertEqual(values["excluded_dlcs"], [])
        self.assertEqual(
            values["binary_path"],
            os.path.join(
                "/",
                "Steam",
                "SteamApps",
                "common",
                "Crusader Kings II",
                "ck2.exe"
            )
        )
        self.assertEqual(
            values["mod_path"],
            os.path.join(
                "/home",
                "Documents",
                "Paradox Interactive",
                "Crusader Kings II",
                "mod"
            )
        )

    @patch("pyparadox.config.sys", autospec=True)
    @patch("pyparadox.config.appdirs.user_data_dir", autospec=True)
    @patch("pyparadox.config.os.path.expanduser", autospec=True)
    def test_default_names_osx(self, config_expanduser, config_user_data_dir,
                               config_sys):
        config_sys.configure_mock(platform="darwin")
        config_user_data_dir.return_value = "Steam"
        config_expanduser.return_value = "/home"

        values = default_values("eu4")
        self.assertEqual(
            values["binary_path"],
            os.path.join(
                "Steam",
                "steamapps",
                "common",
                "Europa Universalis IV",
                "eu4.app"
            )
        )
        self.assertEqual(
            values["mod_path"],
            os.path.join(
                "/home",
                "Documents",
                "Paradox Interactive",
                "Europa Universalis IV",
                "mod"
            )
        )

    @patch("pyparadox.config.sys", autospec=True)
    @patch("pyparadox.config.os.path.expanduser", autospec=True)
    def test_default_names_linux(self, config_expanduser, config_sys):
        config_sys.configure_mock(platform="linux")
        config_expanduser.return_value = "/home"

        values = default_values("ck2")
        self.assertEqual(
            values["binary_path"],
            os.path.join(
                "/home",
                ".steam",
                "root",
                "steamapps",
                "common",
                "Crusader Kings II",
                "ck2"
            )
        )
        self.assertEqual(
            values["mod_path"],
            os.path.join(
                "/home",
                ".paradoxinteractive",
                "Crusader Kings II",
                "mod"
            )
        )


class TestJsonBackend(unittest.TestCase):

    valid_json = {
        '{}': {},
        '{"Foo": "bar"}': {"Foo": "bar"},
        '{"Foo": [1, 2, 3]}': {"Foo": [1, 2, 3]},
        '{"Foo": null}': {"Foo": None}
    }

    def test_read(self):
        for key, value in self.valid_json.items():
            stream = io.StringIO(key)
            self.assertEqual(JsonBackend.read(stream), value)
            self.assertFalse(stream.closed)
            stream.close()

    def test_write(self):
        for key, value in self.valid_json.items():
            stream = io.StringIO()
            JsonBackend.write(value, stream)
            self.assertEqual(re.sub(r"\s+", "", stream.getvalue()),
                             re.sub(r"\s+", "", key))
            self.assertFalse(stream.closed)
            stream.close()


class TestConfig(unittest.TestCase):

    def test_build_config_stream(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}
        with io.StringIO() as stream:
            config = Config.build_config(stream, backend=backend)

        self.assertEqual(dict(config.items()), {"foo": "bar"})

    def test_build_config_file(self):
        config_open = mock_open(read_data="valid json here")
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        with patch("pyparadox.config.open", config_open, create=True):
            config = Config.build_config("/fake/path", backend=backend)

        self.assertEqual(dict(config.items()), {"foo": "bar"})
        config_open.assert_called_once_with("/fake/path")

    def test_build_config_default_file_exists(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}
        with io.StringIO() as stream:
            config = Config.build_config(stream, default={"spam": "eggs"},
                                         backend=backend)

        self.assertEqual(dict(config.items()), {"foo": "bar"})

    def test_build_config_default_file_not_exists(self):
        config_open = mock.Mock()
        config_open.side_effect = FileNotFoundError

        with patch("pyparadox.config.open", config_open, create=True):
            config = Config.build_config("/fake/path", default={"foo": "bar"})

        self.assertEqual(dict(config.items()), {"foo": "bar"})

    def test_load(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        config = Config(backend=backend)
        with io.StringIO() as stream:
            config.load(stream)
            self.assertEqual(dict(config.items()), {"foo": "bar"})

    def test_load_no_override_no_conflict(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        config = Config({"spam": "eggs"}, backend=backend)
        with io.StringIO() as stream:
            config.load(stream, override=False)
            self.assertEqual(sorted(list(config.keys())), ["foo", "spam"])

    def test_load_no_override_with_conflict(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        config = Config({"foo": "spam"}, backend=backend)
        with io.StringIO() as stream:
            with self.assertRaises(AttributeError):
                config.load(stream, override=False)

    def test_load_no_override_with_conflict_suppressed(self):
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        config = Config({"foo": "spam"}, backend=backend)
        with io.StringIO() as stream:
            config.load(stream, override=False, suppress_merge=True)
            self.assertEqual(config["foo"], "spam")

    def test_save(self):
        backend = mock.Mock()

        config = Config(backend=backend)
        stream = mock.Mock()
        config.save(stream)

        backend.write.assert_called_once_with(dict(config.items()), stream)


class TestFileConfig(unittest.TestCase):

    def test_build_config(self):
        config_open = mock_open(read_data="valid json here")
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        with patch("pyparadox.config.open", config_open, create=True):
            config = FileConfig.build_config("/fake/path", backend=backend)

        self.assertEqual(dict(config.items()), {"foo": "bar"})

    def test_load(self):
        config_open = mock_open(read_data="valid json here")
        backend = mock.Mock()
        backend.read.side_effect = lambda stream: {"foo": "bar"}

        config = FileConfig("/fake/path", backend=backend)
        with patch("pyparadox.config.open", config_open, create=True):
            config.load()

        self.assertEqual(dict(config.items()), {"foo": "bar"})

    @patch("pyparadox.config.os.makedirs", autospec=True)
    def test_save(self, config_makedirs):
        config_open = mock_open()
        backend = mock.Mock()

        config = FileConfig("/fake/path", backend=backend)
        with patch("pyparadox.config.open", config_open, create=True):
            config.save()

        backend.write.assert_called_once_with(config, config_open())
        config_makedirs.assert_called_once_with("/fake")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

# from pyparadox import main


class TestMain(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_main(self):
        pass


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess
import logging
import glob
import re

GAMES = {
    "ck2": "Crusader Kings II",
    "eu4": "Europa Universalis IV",
    "stellaris": "Stellaris",
}
# Maybe move this to a user-editable file at some point?  Hard-coding sucks.
EXPANSION_NAMES = [
    # Crusader Kings 2
    "The Sword of Islam",
    "Legacy of Rome",
    "Sunset Invasion",
    "The Republic",
    "The Old Gods",
    "Sons of Abraham",
    "Rajas of India",
    "Charlemagne",
    "Way of Life",
    "Horse Lords",
    "Conclave",
    # Europa Universalis IV
    "Conquest of Paradise",
    "Wealth of Nations",
    "Res Publica",
    "Art of War",
    "El Dorado",
    "Common Sense",
    "The Cossacks",
]

_logger = logging.getLogger(__name__)


class Plugin(object):
    """Container class that represents a single plugin (mod/DLC/expansion)."""

    MOD = 0
    DLC = 1
    EXPANSION = 2

    def __init__(self, name, file_name, plugin_type=0, enabled=True):
        """:param str name: Full name of the plugin.
        :param str file_name: Name of the plugin file.
        :param int plugin_type: Integer that determines the plugin type.
        :param bool enabled: Whether or not the plugin is currently selected
                             for use.
        """
        self.name = name
        self.file_name = file_name
        self.plugin_type = plugin_type
        self.enabled = enabled


def find_plugins(path, extension, config=None):
    """Find plugins in *path* that match *extension*.

    If *config* is provided, it will determine whether the plugins should be
    enabled or disabled on instantiation.  By default, the mods will be
    disabled and the DLCs will be enabled.

    :param str path: File system path to the directory containing the plugins.
    :param str extension: File type extension of the plugins.  Needn't begin
                          with a full stop, but it's optional.
    :keyword config: Configuration mapping that determines the *enabled* status
                     of the generated plugins.
    :type config: :class:`Config <pyparadox.config.Config>` or :class:`dict`
    :return: List of plugins.
    :rtype: [:class:`Plugin`]
    """
    extension = extension.strip(".")
    plugin_paths = glob.glob(os.path.join(path, "*.{}".format(extension)))

    plugins = []

    if config is not None:
        mods = config["mods"]
        excluded_dlcs = config["excluded_dlcs"]
    else:
        mods = []
        excluded_dlcs = []

    for path in plugin_paths:
        file_name = os.path.basename(path)
        _logger.debug("found {}".format(path))

        with open(path) as fp:
            try:
                name = find_plugin_name(fp.read())
            except ValueError:
                name = file_name

        if extension == "mod":
            plugin_type = Plugin.MOD
            enabled = file_name in mods
        elif extension == "dlc":
            plugin_type = (Plugin.EXPANSION if name in EXPANSION_NAMES else
                           Plugin.DLC)
            enabled = file_name not in excluded_dlcs
        else:
            # Assume it is a mod. This technically should never happen.
            _logger.error("{} does not have the correct file extension".format(
                path))
            plugin_type = Plugin.MOD
            enabled = file_name in mods

        plugin = Plugin(name, file_name, plugin_type=plugin_type,
                        enabled=enabled)
        plugins.append(plugin)

    return plugins


def find_plugin_name(plugin_contents):
    """Find the proper name of a plugin inside the contents of its file.

    :param str plugin_contents: The entire contents of the plugin file (or just
                                the part that contains the necessary
                                information).
    :return: Proper name of the plugin.
    :rtype: str
    :raise ValueError: If a plugin name couldn't be found.
    """
    try:
        return re.search('^name[ \t]*=[ \t]*"(.*)"', plugin_contents,
                         re.MULTILINE).group(1)
    except AttributeError as e:
        raise ValueError("Could not find plugin name",
                         plugin_contents) from e


def make_command(binary_path, mods, excluded_dlcs):
    """Compile a command that can be parsed by the :mod:`subprocess` module.

    :param str binary_path: Path (absolute) towards the executable.
    :param list mods: List of mods to be enabled.
    :param list excluded_dlcs: List of DLCs to be disabled.
    :return: Command that the :mod:`subprocess` module can process.
    :rtype: list
    """
    command = [binary_path]
    command.append("-skiplauncher")
    for mod in mods:
        command.append("-mod={}".format(os.path.join("mod", mod)))
    for dlc in excluded_dlcs:
        command.append("-exclude_dlc={}".format(os.path.join("dlc", dlc)))
    if sys.platform == "darwin":
        command.insert(1, "--args")
        command = ["open", "-n"] + command
    return command


def execute_command(command, synchronous=False):
    """Execute the command in a separate process.

    If *synchronous* is :const:`True`, the program will wait until the other
    process has finished.

    :param list command: Command to be executed.
    :param bool synchronous: Whether or not the subprocess should be
                             synchronous.
    :return: The spawned process.
    """
    _logger.debug("executing {}".format(" ".join(command)))
    if synchronous:
        return subprocess.call(command)
    else:
        return subprocess.Popen(command)



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import textwrap
import unittest
from unittest import mock

from pyparadox import core

_CURRENT_DIRECTORY = os.path.dirname(__file__)
_MOD_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-mod")
_GAME_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-game")
_DLC_DIRECTORY = os.path.join(_GAME_DIRECTORY, "dlc")
_CONFIG_DIRECTORY = os.path.join(_CURRENT_DIRECTORY, "fake-config")


class TestCore(unittest.TestCase):

    def test_find_plugins_mod(self):
        plugins = core.find_plugins(_MOD_DIRECTORY, "mod")

        self.assertEqual(len(plugins), 1)

        plugin = plugins[0]

        self.assertEqual(plugin.name, "Better gender law mod")
        self.assertEqual(plugin.file_name, "better gender law mod.mod")
        self.assertEqual(plugin.plugin_type, core.Plugin.MOD)
        self.assertEqual(plugin.enabled, False)

    def test_find_plugins_mod_enabled(self):
        config_dict = {"mods": ["better gender law mod.mod"]}
        config = mock.MagicMock()
        config.__getitem__.side_effect = lambda key: config_dict.get(key)

        plugins = core.find_plugins(_MOD_DIRECTORY, "mod", config=config)

        plugin = plugins[0]

        self.assertTrue(plugin.enabled)

    def test_find_plugins_mod_disabled(self):
        config_dict = {"mods": []}
        config = mock.MagicMock()
        config.__getitem__.side_effect = lambda key: config_dict.get(key)

        plugins = core.find_plugins(_MOD_DIRECTORY, "mod", config=config)

        plugin = plugins[0]

        self.assertFalse(plugin.enabled)

    def test_find_plugins_dlc(self):
        plugins = core.find_plugins(_DLC_DIRECTORY, "dlc")

        self.assertEqual(len(plugins), 2)

        plugins = sorted(plugins, key=lambda x: x.file_name)

        dlc = plugins[0]
        expansion = plugins[1]

        self.assertEqual(dlc.name, "Dynasty Coat of Arms Pack 1")
        self.assertEqual(dlc.file_name, "dlc001.dlc")
        self.assertEqual(dlc.plugin_type, core.Plugin.DLC)
        self.assertEqual(dlc.enabled, True)

        self.assertEqual(expansion.name, "Way of Life")
        self.assertEqual(expansion.file_name, "dlc050.dlc")
        self.assertEqual(expansion.plugin_type, core.Plugin.EXPANSION)
        self.assertEqual(expansion.enabled, True)

    def test_find_plugins_dlc_enabled(self):
        config_dict = {"excluded_dlcs": []}
        config = mock.MagicMock()
        config.__getitem__.side_effect = lambda key: config_dict.get(key)

        plugins = core.find_plugins(_DLC_DIRECTORY, "dlc", config=config)

        for plugin in plugins:
            self.assertTrue(plugin.enabled)

    def test_find_plugins_dlc_disabled(self):
        config_dict = {"excluded_dlcs": ["dlc001.dlc", "dlc050.dlc"]}
        config = mock.MagicMock()
        config.__getitem__.side_effect = lambda key: config_dict.get(key)

        plugins = core.find_plugins(_DLC_DIRECTORY, "dlc", config=config)

        for plugin in plugins:
            self.assertFalse(plugin.enabled)

    def test_find_plugin_name(self):
        string = textwrap.dedent("""\
            name = "foo"
            archive = "bar/bar42.zip"
            steam_id = "1337"
            gamersgate_id = "404"
            checksum = "longstringoftext"
            affects_checksum = no""")

        name = core.find_plugin_name(string)

        self.assertEqual(name, "foo")

    def test_find_plugin_name_not_found(self):
        string = textwrap.dedent("""\
            naaaaaame = "foo"
            archive = "bar/bar42.zip"
            steam_id = "1337"
            gamersgate_id = "404"
            checksum = "longstringoftext"
            affects_checksum = no""")

        with self.assertRaises(ValueError):
            core.find_plugin_name(string)

    def test_make_command(self):
        binary_path = "bin"
        mods = ["mod1", "mod2"]
        excluded_dlcs = ["dlc1", "dlc2"]

        command = core.make_command(binary_path, mods, excluded_dlcs)

        expected = ["bin", "-skiplauncher", "-mod=mod/mod1", "-mod=mod/mod2",
                    "-exclude_dlc=dlc/dlc1", "-exclude_dlc=dlc/dlc2"]

        self.assertEqual(command, expected)

    def test_make_command_plain(self):
        binary_path = "bin"

        command = core.make_command(binary_path, [], [])

        expected = ["bin", "-skiplauncher"]

        self.assertEqual(command, expected)

    def test_execute_command(self):
        start = time.time()

        process = core.execute_command(
            [os.path.join(_GAME_DIRECTORY, "game")])

        end = time.time()
        diff = end - start

        self.assertLess(diff, 0.1)
        self.assertTrue(process)

    def test_execute_command_synchronous(self):
        start = time.time()

        output = core.execute_command(
            [os.path.join(_GAME_DIRECTORY, "game")],
            synchronous=True)

        end = time.time()
        diff = end - start

        self.assertGreaterEqual(diff, 0.1)
        self.assertEqual(output, 0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import argparse
import logging
import logging.config
import json
from pkg_resources import resource_filename

import appdirs

from .config import default_values, FileConfig, DEFAULT_CONFIG_DIR
from .core import GAMES, make_command, execute_command
from . import __version__

_logger = logging.getLogger(__name__)


def main_console(args=sys.argv):
    parser_args = _parse_args(args=args)
    _setup_logging(parser_args.loglevel)
    _logger.info("pyparadox launched")
    config = _create_config(parser_args.game)

    command = make_command(config["binary_path"],
                           config["excluded_dlcs"],
                           config["mods"])
    execute_command(command)
    _logger.info("shutting down pyparadox")


def main_gui(args=sys.argv):
    parser_args = _parse_args(args=args)
    _setup_logging(parser_args.loglevel)
    _logger.info("pyparadox-qt launched")
    config = _create_config(parser_args.game)

    _logger.debug("loading PyQt5")
    from PyQt5 import QtWidgets
    from .gui import MainWindow

    _logger.debug("creating QApplication")
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationVersion(__version__)
    app.setApplicationName("PyParadox")

    _logger.debug("creating MainWindow")
    form = MainWindow(config)
    form.show()

    _logger.debug("starting Qt exec loop")
    return_code = app.exec_()

    _logger.info("shutting down pyparadox-qt")
    sys.exit(return_code)


def main_qml(args=sys.argv):
    parser_args = _parse_args(args=args)
    _setup_logging(parser_args.loglevel)
    _logger.info("pyparadox-qml launched")
    config = _create_config(parser_args.game)
    game = parser_args.game

    _logger.debug("loading PyQt5")
    from PyQt5 import QtCore, QtGui, QtWidgets, QtQml
    from .gui import PluginListModel, LogicWrapper

    _logger.debug("creating QAapplication")
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationVersion(__version__)
    app.setApplicationName("PyParadox")

    _logger.debug("creating QQmlApplicationEngine")
    engine = QtQml.QQmlApplicationEngine()

    rootContext = engine.rootContext()

    modsModel = PluginListModel()
    dlcsModel = PluginListModel()
    expansionsModel = PluginListModel()

    logicWrapper = LogicWrapper(config, modsModel, dlcsModel, expansionsModel)

    rootContext.setContextProperty("imagePath", resource_filename(
        "pyparadox",
        "resources/{}.png".format(game)))
    rootContext.setContextProperty("modsModel", modsModel)
    rootContext.setContextProperty("dlcsModel", dlcsModel)
    rootContext.setContextProperty("expansionsModel", expansionsModel)
    rootContext.setContextProperty("logicWrapper", logicWrapper)

    _logger.debug("loading main.qml")
    engine.load(QtCore.QUrl(resource_filename(
        "pyparadox",
        "resources/ui/main.qml")))
    window = engine.rootObjects()[0]
    window.setIcon(QtGui.QIcon(resource_filename(
        "pyparadox",
        "resources/paradox.png")))
    _logger.debug("starting Qt exec loop")
    return_code = app.exec_()

    _logger.info("shutting down pyparadox-qml")
    sys.exit(return_code)


def _parse_args(args=sys.argv):
    if args == sys.argv:
        args = args[1:]
    parser = argparse.ArgumentParser(description="Launcher for Paradox titles")
    parser.add_argument("game", type=str, choices=list(GAMES.keys()),
                        help="Game to launch")
    parser.add_argument("-v", "--verbose", action="store_const",
                        dest="loglevel", const=logging.INFO)
    parser.add_argument("-d", "--debug", action="store_const", dest="loglevel",
                        const=logging.DEBUG)
    parser.add_argument("steam-command", nargs="?", help=argparse.SUPPRESS)

    return parser.parse_args(args)


def _setup_logging(loglevel=None):

    with open(resource_filename("pyparadox", "resources/logging.json")) as fp:
        config = json.load(fp)

    if loglevel is not None:
        config["handlers"]["console"]["level"] = loglevel

    log_dir = appdirs.user_data_dir("PyParadox")
    try:
        os.makedirs(log_dir)
    except FileExistsError:
        pass

    default_file = config["handlers"]["info_file_handler"]["filename"]
    config["handlers"]["info_file_handler"]["filename"] = \
        os.path.join(log_dir, default_file)

    logging.config.dictConfig(config)

    _logger.debug("logging initiated")


def _create_config(game):
    path = os.path.join(DEFAULT_CONFIG_DIR, "pyparadox_{}.json".format(game))

    _logger.info("loading config file from {}".format(path))
    config = FileConfig.build_config(path, default=default_values(game))

    return config


if __name__ == "__main__":
    main_gui()



# -*- coding: utf-8 -*-
#
# PyParadox is a nix launcher for Paradox titles.
# Copyright (C) 2014  Carmen Bianca Bakker <carmenbbakker@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest import mock
from unittest.mock import patch

from PyQt5 import QtWidgets
from PyQt5.QtTest import QTest
from PyQt5.QtCore import Qt

from pyparadox import gui


class TestMainWindow(unittest.TestCase):

    @patch('pyparadox.gui.LogicWrapper')
    def setUp(self, gui_LogicWrapper):
        self.app = QtWidgets.QApplication([])
        self.config = mock.MagicMock()
        self.form = gui.MainWindow(self.config)
        # Does not verify whether only legal LogicWrapper methods are called.
        # Autospeccing results in problems with Qt slots and signals.
        self.assertTrue(self.form._logicWrapper.resetModels.called)
        self.form._logicWrapper.reset_mock()

    def test_on_runBtn_clicked(self):
        logicWrapper = self.form._logicWrapper
        QTest.mouseClick(self.form.runBtn, Qt.LeftButton)
        self.assertTrue(logicWrapper.savePlugins.called)
        self.assertTrue(logicWrapper.run.called)

    @patch('pyparadox.gui.ConfigDlg')
    def test_on_configBtn_clicked_no_save(self, gui_ConfigDlg):
        configDlg = gui_ConfigDlg()
        configDlg.exec_.return_value = False
        QTest.mouseClick(self.form.configBtn, Qt.LeftButton)
        self.assertTrue(configDlg.exec_.called)
        self.assertFalse(self.form._logicWrapper.savePlugins.called)
        self.assertFalse(self.form._logicWrapper.resetModels.called)

    @patch('pyparadox.gui.ConfigDlg')
    def test_on_configBtn_clicked_save(self, gui_ConfigDlg):
        configDlg = gui_ConfigDlg()
        configDlg.exec_.return_value = True
        configDlg.binaryPathEdit.text.return_value = "foo"
        configDlg.modPathEdit.text.return_value = "bar"
        QTest.mouseClick(self.form.configBtn, Qt.LeftButton)
        self.assertTrue(configDlg.exec_.called)
        self.form._config.__setitem__.assert_any_call("binary_path", "foo")
        self.form._config.__setitem__.assert_any_call("mod_path", "bar")
        self.assertTrue(self.form._logicWrapper.savePlugins.called)
        self.assertTrue(self.form._logicWrapper.resetModels.called)


class TestMainWindowSlots(unittest.TestCase):

    @patch('pyparadox.gui.find_plugins')
    def setUp(self, gui_find_plugins):
        self.app = QtWidgets.QApplication([])
        self.config = mock.MagicMock()
        self.form = gui.MainWindow(self.config)

    @patch('pyparadox.gui.QtWidgets.QMessageBox', autospec=True)
    def test_runFailed(self, gui_QMessageBox):
        self.form._logicWrapper.runFailed.emit("foo", "bar")
        self.assertTrue(gui_QMessageBox.called)
        msgBox = gui_QMessageBox()
        self.assertTrue(msgBox.exec_.called)

    def test_runSucceeded(self):
        self.form._logicWrapper.runSucceeded.emit()
        # TODO: Figure out how to detect whether the window was successfully
        # closed.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.version_info < (3, 3):
    raise RuntimeError('PyParadox requires Python >= 3.3')


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    "appdirs"
]

test_requirements = [
]

# Copied from (and hacked):
# https://github.com/pypa/virtualenv/blob/develop/setup.py#L42
def get_version(filename):
    import re

    here = os.path.dirname(__file__)
    with open(os.path.join(here, filename)) as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)

setup(
    name='pyparadox',
    version=get_version(os.path.join("pyparadox", "__init__.py")),
    description="PyParadox is a nix launcher for Paradox titles.",
    long_description=readme + '\n\n' + history,
    author="Carmen Bianca Bakker",
    author_email='carmenbbakker@gmail.com',
    url='https://gitlab.com/carmenbbakker/pyparadox',
    packages=[
        'pyparadox',
    ],
    package_dir={'pyparadox': 'pyparadox'},
    entry_points={
        'gui_scripts': [
            'pyparadox-qt = pyparadox.main:main_gui',
            'pyparadox-qml = pyparadox.main:main_qml'
        ],
        'console_scripts': [
            'pyparadox = pyparadox.main:main_console'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="GPLv3+",
    zip_safe=False,
    keywords='pyparadox',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Games/Entertainment',
        'Topic :: Utilities',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
